#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "pcd8544.h"
#include <stdlib.h>
#include "snake.h"

coord_t snake[NMAX];
int snake_size;
int snake_dir;
coord_t feed;
int show_score = 0;
byte buffer[20];

int check_food_pos()
{
    if(feed.x == snake[0].x + oriz[snake_dir] && feed.y == snake[0].y + vert[snake_dir])
    {
        return 0;
    }
    
    int i;
    for(i = 0; i < snake_size; ++i)
    {
        if(feed.x == snake[i].x && feed.y == snake[i].y)
            return 0;
    }
    
    return 1;
}

void generate_food()
{
    do
    {
        feed.x = rand() % 30 + 5;
        feed.y = rand() % 30 + 5;
        
        feed.x -= feed.x % 2;
        feed.y -= feed.y % 2;
    }
    while(check_food_pos());
}

void checkButton(int *direction) {

	if ((PIND & (1 << LEFT_BUTTON)) == 0) {
		*direction = LEFT;
		return;
	}
	
	if ((PIND & (1 << UP_BUTTON)) == 0) {
		*direction = UP;
		return;
	}
	
	if ((PIND & (1 << RIGHT_BUTTON)) == 0) {
		*direction = RIGHT;
		return;
	}
	
	if ((PIND & (1 << DOWN_BUTTON)) == 0) {
		*direction = DOWN;
		return;
	}
}


// Snake structure initialization.
void initSnake() {
	int i;

	snake[0].x = 10;
	snake[0].y = 10;
	snake_size = 4;
	snake_dir = RIGHT;
	
	for (i = 1; i < snake_size; i++) {
		snake[i].x = snake[i - 1].x + oriz[RIGHT];
		snake[i].y = snake[i - 1].y + vert[RIGHT];
	}
    
    int garbage;
	
	srand(garbage);
    
	generate_food();
	
	show_score = 0;
}

// Check if snake died.
int checkSnake() {
	int i, j;
	
	for (i = 0; i < snake_size; i++)
		if (snake[i].x <= 1 || snake[i].x >= LCD_X_RES ||
		    snake[i].y <= 1 || snake[i].y >= LCD_Y_RES)
			return DEAD;
			
	for (i = 0; i < snake_size; i++)
		for (j = i + 1; j < snake_size; j++)
		if (snake[i].x == snake[j].x && snake[i].y == snake[j].y)
			return DEAD;
	
	if (snake[snake_size - 1].x == feed.x && 
	    snake[snake_size - 1].y == feed.y) {
		
		snake[snake_size].x = snake[snake_size - 1].x + oriz[snake_dir];
		snake[snake_size].y = snake[snake_size - 1].y + vert[snake_dir];
		snake_size++;
		feed.x = 0;
		show_score = 1;
	}
	
	return ALIVE;
}

void clearSnake() {
	int i;
	
	for (i = 0; i < snake_size; i++) 
		LcdPixel(snake[i].x, snake[i].y, PIXEL_OFF);
		
	if (feed.x != 0) 
		LcdPixel(feed.x, feed.y, PIXEL_OFF);
}

void showSnake() {
	int i;
	
	for (i = 0; i < snake_size; i++)
    {
		LcdPixel(snake[i].x, snake[i].y, PIXEL_ON);
        LcdPixel(snake[i].x + 1, snake[i].y, PIXEL_ON);
        LcdPixel(snake[i].x, snake[i].y + 1, PIXEL_ON);
        LcdPixel(snake[i].x + 1, snake[i].y + 1, PIXEL_ON);
    }
		
	if (feed.x != 0) 
    {
		LcdPixel(feed.x, feed.y, PIXEL_ON);
        LcdPixel(feed.x + 1, feed.y, PIXEL_ON);
        LcdPixel(feed.x, feed.y + 1, PIXEL_ON);
        LcdPixel(feed.x + 1, feed.y + 1, PIXEL_ON);
    }
}

void moveSnake(int snake_dir) {
	int i;
	coord_t tmp, tmp1;
	
	if (snake_dir == -1) return;

	tmp.x = snake[snake_size - 1].x + oriz[snake_dir];
	tmp.y = snake[snake_size - 1].y + vert[snake_dir];
	
	for (i = snake_size - 1; i >= 0; i--) {
		tmp1 = snake[i];
		snake[i] = tmp;
		tmp = tmp1;
	}

}

int checkDirection(int old, int current) {
	if (old == RIGHT && (current == RIGHT || current == LEFT)) return 0;
	if (old == LEFT  && (current == LEFT || current == RIGHT)) return 0;
	if (old == UP && (current == UP || current == DOWN)) return 0;
	if (old == DOWN && (current == DOWN || current == UP)) return 0;
	return 1;
 }


void snake_main() {

	int i, dir, dead = 0;

	dead = 0;
	show_score = 0;
	initSnake();
	

    LcdContrast(0x7F);
    LcdClear();
    LcdGotoXYFont(2,3);
    LcdStr(FONT_2X, "SNAKE");
    LcdUpdate();
    _delay_ms(2500);
	LcdClear();

	showSnake();
		
 	snake_dir = RIGHT;
	while (!dead) {
		LcdClear();
		dir = -1;
        
        _delay_ms(50);
        
		for (i = 0; i < 80; i++) {
			checkButton(&dir);
			if (dir != -1) break;
		}
		
		clearSnake();
		if (dir != -1 && checkDirection(snake_dir, dir)) {
			moveSnake(dir);
			snake_dir = dir;
		} else {
			moveSnake(snake_dir);
		}
	
		dead = checkSnake();
		if (dead) continue;
		showSnake();
		
		LcdUpdate();		
		
		if (feed.x == 0) {
			feed.x = rand() % (LCD_X_RES - 10) + 4;
			feed.y = rand() % (LCD_Y_RES - 10) + 4;
            
            feed.x -= feed.x % 2;
            feed.y -= feed.y % 2;
		}
	}
	
	LcdClear();
    LcdGotoXYFont(2,3);
    LcdStr(FONT_1X, "GAME OVER!");
    LcdUpdate();
	_delay_ms(500);
	
	dir = -1;
	do {
		checkButton(&dir);		
	} while (dir == -1);
}


