#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "pcd8544.h"
#include <stdlib.h>
#include "tetris.h"

#define LEFT_BUTTON      PD1
#define DOWN_BUTTON      PD0
#define RIGHT_BUTTON     PD3
#define ROTATE_BUTTON    PD2

unsigned short int table[TABLE_SIZE_Y];
unsigned short int marked_lines;

unsigned short int score;

void set_piece_offsets(piece_t *p, piece_type t)
{
    switch(t)
    {
        case O:
            p->segment_offset_x[0] = 1;
            p->segment_offset_x[1] = 1;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 0;
            p->right_clearance = 1;
            
            break;
            
        case I_1:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 2;
            p->segment_offset_y[2] = 3;
            
            p->left_clearance = 0;
            p->right_clearance = 0;
            
            break;
        
        case I_2:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 1;
            p->segment_offset_x[2] = 2;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 0;
            p->segment_offset_y[2] = 0;
            
            p->left_clearance = 1;
            p->right_clearance = 2;
            
            break;
            
        case S_1:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = -1;
            p->segment_offset_x[2] = -1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 1;
            p->right_clearance = 0;
            
            break;
            
        case S_2:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
        
        case Z_1:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 1;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 0;
            p->right_clearance = 1;
            
            break;
        
        case Z_2:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 0;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
        
        case J_1:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 1;
            p->right_clearance = 0;
            
            break;
        
        case J_2:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = -1;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 0;
            p->segment_offset_y[2] = 0;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
            
        case J_3:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 2;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 0;
            p->right_clearance = 1;
            
            break;
            
        case J_4:
            p->segment_offset_x[0] = -2;
            p->segment_offset_x[1] = -1;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 2;
            p->right_clearance = 0;
            
            break;
            
        case L_1:
            p->segment_offset_x[0] = 1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 0;
            p->right_clearance = 1;
            
            break;
            
        case L_2:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 1;
            p->segment_offset_x[2] = 2;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 0;
            p->right_clearance = 2;
            
            break;
            
        case L_3:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = -1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 2;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 1;
            p->right_clearance = 0;
            
            break;
        
        case L_4:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 1;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 0;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
            
        case T_1:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 0;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 0;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
        
        case T_2:
            p->segment_offset_x[0] = 0;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 2;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 0;
            p->right_clearance = 1;
            
            break;
        
        case T_3:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 1;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 1;
            
            p->left_clearance = 1;
            p->right_clearance = 1;
            
            break;
        
        case T_4:
            p->segment_offset_x[0] = -1;
            p->segment_offset_x[1] = 0;
            p->segment_offset_x[2] = 0;
            
            p->segment_offset_y[0] = 1;
            p->segment_offset_y[1] = 1;
            p->segment_offset_y[2] = 2;
            
            p->left_clearance = 1;
            p->right_clearance = 0;
            
            break;
    }
}

void draw_square3x3(sbyte x, sbyte y)
{
    sbyte i, j;
    for(i = 0; i < 3; ++i)
    {
        for(j = 0; j < 3; ++j)
            LcdPixel(x + i, y + j, PIXEL_ON);
    }
}

void draw_border()
{
    sbyte i;
    for(i = 0; i < LCD_X_RES; ++i)
    {
        LcdPixel(i, 0, PIXEL_ON);
        LcdPixel(i, 1, PIXEL_ON);
        LcdPixel(i, 32, PIXEL_ON);
        LcdPixel(i, 33, PIXEL_ON);
    }
}

void draw_piece(piece_t *p)
{
    draw_square3x3(p->y * 3, p->x * 3 + 2);
    
    sbyte i;
    for(i = 0; i < 3; ++i)
    {
        draw_square3x3((p->y + p->segment_offset_y[i]) * 3, (p->x + p->segment_offset_x[i]) * 3 + 2);
    }
}

void init_table()
{
    sbyte i, j;
    
    for(i = 0; i < TABLE_SIZE_X; ++i)
    {
        table[i] = 0;
    }
}

void draw_table()
{
    sbyte i, j;
    
    for(i = 0; i < TABLE_SIZE_X; ++i)
    {
        for(j = 0; j < TABLE_SIZE_Y; ++j)
        {
            if(table[j] & 1<<i)
                draw_square3x3(j * 3, i * 3 + 2);
        }
    }
}

sbyte check_position(piece_t *p)
{
	if(p->x < 0 || p->x >= TABLE_SIZE_X)
		return 0;
        
    if(table[p->y] & 1<<p->x)
        return 0;
    
	sbyte i;
	for(i = 0; i < 3; ++i)
	{
		if((p->x + p->segment_offset_x[i]) < 0 || (p->x + p->segment_offset_x[i]) >= TABLE_SIZE_X)
			return 0;
        
        if(table[p->y + p->segment_offset_y[i]] & 1<<(p->x + p->segment_offset_x[i]))
            return 0;
            
        /*LcdClear();
        LcdGotoXYFont(1,1);
        char buf[15];
        sprintf(buf, "%d %d %d %d", p->x, p->y, p->segment_offset_x[i], p->segment_offset_y[i]);
        LcdStr(FONT_1X, buf);
        LcdUpdate();
        _delay_ms(500);*/
	}

	return 1;
}

void try_rotate(piece_t *p, piece_type rotation)
{
	piece_t new_piece = *p;
	
    new_piece.type = rotation;
	set_piece_offsets(&new_piece, rotation);

	sbyte i;
	if(check_position(&new_piece)) {*p = new_piece; return; }
	
	new_piece.x = p->x + 1;
    if(check_position(&new_piece)) {*p = new_piece; return; }

    new_piece.x = p->x - 1;
    if(check_position(&new_piece)) {*p = new_piece; return; }		
}

void rotate(piece_t *p)
{
	if(p->type == O) return;
	
	if(p->type >= I_1)
	{
		try_rotate(p, (piece_type)((p->type % 2)?(p->type - 1):(p->type + 1)));
		return;
	}
	
	try_rotate(p, (piece_type)((p->type % 4 == 3)?(p->type - 3):(p->type + 1)));	
}

sbyte can_advance(piece_t *p)
{
	piece_t new_pos = *p;
	new_pos.y--;
    
    if(new_pos.y < 0) return 0;

	if(check_position(&new_pos)) return 1;
	else return 0;
    
    return 1;
}

void try_move(piece_t *p, sbyte offset)
{
	piece_t new_pos = *p;
	new_pos.x += offset;
	
	if(check_position(&new_pos)) *p = new_pos;
}

void copy_to_table(piece_t *p)
{
	table[p->y] |= 1<<p->x;
	
	sbyte i;
	for(i = 0; i < 3; ++i)
	{
		table[p->y + p->segment_offset_y[i]] |= 1<<(p->x + p->segment_offset_x[i]);
	}
}

void check_button(button_t *button) {

	if ((PIND & (1 << LEFT_BUTTON)) == 0) {
		*button = LEFT;
		return;
	}
	
	if ((PIND & (1 << ROTATE_BUTTON)) == 0) {
		*button = ROTATE;
		return;
	}
	
	if ((PIND & (1 << RIGHT_BUTTON)) == 0) {
		*button = RIGHT;
		return;
	}
	
	if ((PIND & (1 << DOWN_BUTTON)) == 0) {
		*button = DOWN;
		return;
	}
}

void check_lines()
{
    marked_lines = 0;
    
    sbyte i;
    for(i = 0; i < TABLE_SIZE_Y; ++i)
    {
        if(table[i] == FULL_LINE)
        {
            marked_lines |= 1<<i;
            score++;
        }
        
        if(table[i] == 0)
            return;
    }
}

void remove_lines()
{
    sbyte i = 0, offset = 0;
    
    while(i + offset < TABLE_SIZE_Y)
    {
        table[i] = table[i + offset];
        
        if(marked_lines & 1)
        {
            ++offset;
        }        
        else
        {
            ++i;
        }
        
        marked_lines >>= 1;
    }
    
    for(;i < TABLE_SIZE_Y; ++i)
    {
        table[i] = 0;
    }
}

void tetris_main()
{
    srand(1);    
    
    init_table();
    
    LcdContrast(0x7F);
    LcdClear();
    LcdGotoXYFont(2,3);
    LcdStr(FONT_2X, "TETRIS");
    LcdUpdate();
    _delay_ms(500);
    LcdClear();
    
    char buf[10];
    
    TCCR1B |= 1 << CS10; //initializare timer cu Fcpu/64
    
    sbyte game_over = 0, spawned = 0;
    button_t button = -1, last_button = -1;
    
    unsigned int speed = 15;
    
    piece_t current, next;
    
    next.type = (piece_type)(rand()%NUM_TYPES);
    set_piece_offsets(&next, next.type);
    
    while(!game_over)
    {
        
        
        if(!spawned)
        {
            current = next;
            next.type = (piece_type)(rand()%NUM_TYPES);
            set_piece_offsets(&next, next.type);
            next.x = 11 + next.left_clearance;
            next.y = 20;
            
            current.y = TABLE_SIZE_Y;
            current.x = TABLE_SIZE_X / 2;
            
            if(!can_advance(&current)) break;
            
            spawned = 1;
        }
        
        int i;
        for(i = 0; i < speed;)
        {
            check_button(&button);
            
            if (TCNT1 >= 15000) 
            {        
                
                if(last_button != button) switch(button)
                {
                    case ROTATE:
                        rotate(&current);
                        break;
                    case LEFT:
                        try_move(&current, -1);                        
                        break;
                    case RIGHT:
                        try_move(&current, 1);
                        break;
                    case DOWN:
                        while(can_advance(&current))
                        {
                            current.y--;
                        }
                        i = speed;
                        break;
                }
                
                if(button != -1 && button != DOWN)
                {
                    LcdClear();
                    draw_border();
                    LcdGotoXYFont(2,6);
                    sprintf(buf, "%d", score);
                    LcdStr(FONT_1X, buf);
                    draw_piece(&current);
                    draw_piece(&next);                        
                    draw_table();                        
                    LcdUpdate();
                }
                
                last_button = button;                
                button = -1;
                TCNT1 = 0;
                i++;
            }
        }
        
        if(can_advance(&current))
        {
            current.y--;
            /*LcdClear();
            LcdGotoXYFont(1,1);
            char buf[15];
            sprintf(buf, "%d %d", current.x, current.y);
            LcdStr(FONT_1X, buf);*/
        }
        else        
        {
            spawned = 0; 
            copy_to_table(&current);
            
            check_lines();
            remove_lines();

            marked_lines = 0;

            speed = 15 - score / 5;
            if(speed < 3) speed = 3;
        }
        
        LcdClear();
        draw_border();
        if(spawned) draw_piece(&current);
        draw_piece(&next);
        LcdGotoXYFont(2,6);
        sprintf(buf, "%d", score);
        LcdStr(FONT_1X, buf);
        draw_table();
        
        LcdUpdate();
    }
    
    TCCR1B &= ~(1 << CS10);
    
    LcdClear();
    LcdGotoXYFont(2,3);
    LcdStr(FONT_1X, "GAME OVER!");
    LcdUpdate();
	_delay_ms(500);
	
	button = -1;
	do {
		check_button(&button);		
	} while (button == -1);
}